-- compute
#version 450 core

// This shader implements a sorting network for 1024 elements.
//
// It is follows the alternative notation for bitonic sorting networks, as given at:
// https://en.m.wikipedia.org/wiki/Bitonic_sorter#Alternative_representation

#define LocalBitonicMergeSort	0
#define LocalDisperse           1
#define BigFlip                 2
#define BigDisperse             3

layout ( local_size_x = 512 ) in;

layout ( binding = 0 ) buffer SortData 
{
	uint value [];
};

uniform int step;
uniform int algorithm;

shared uint local_value [gl_WorkGroupSize.x * 2];

bool is_smaller ( in const uint left, in const uint right )
{
	return left > right;
}

#define COMPARE is_smaller

void globalCompareAndSwap ( ivec2 idx )
{
	if ( COMPARE ( value [idx.x], value [idx.y] ) )
	{
		uint tmp = value [idx.x];

		value [idx.x] = value [idx.y];
		value [idx.y] = tmp;
	}
}

void localCompareAndSwap ( ivec2 idx )
{
	if ( COMPARE ( local_value [idx.x], local_value [idx.y] ) )
	{
		uint tmp = local_value [idx.x];

		local_value [idx.x] = local_value [idx.y];
		local_value [idx.y] = tmp;
	}
}

	// Performs full-height flip (h height) over globally available indices.
void bigFlip ( in uint h )
{
	uint tPrime = gl_GlobalInvocationID.x;
	uint halfH  = h >> 1;
	uint q      = ((2 * tPrime) / h) * h;
	uint t2     = tPrime % halfH;
	uint x      = q + t2;
	uint y      = q + h - t2 - 1; 

	globalCompareAndSwap ( ivec2 ( x, y ) );
}

	// full-height disperse (h height) over globally available indices.
void bigDisperse ( in uint h )
{
	uint tPrime = gl_GlobalInvocationID.x;
	uint halfH  = h >> 1;
	uint q      = ((2 * tPrime) / h) * h;
	uint t2     = tPrime % halfH;
	uint x      = q + t2;
	uint y      = q + t2 + halfH;

	globalCompareAndSwap ( ivec2 ( x, y ) );
}

	// Performs full-height flip (h height) over locally available indices.
void localFlip ( in uint h )
{
	uint	t = gl_LocalInvocationID.x;
	uint	halfH  = h >> 1;
	uint	t2     = t % halfH;
	uint	q      = h * ( ( 2 * t ) / h );

	barrier ();

	localCompareAndSwap ( ivec2 ( q + t2, q + h - 1 - t2 ) );
}

// Performs progressively diminishing disperse operations (starting with height h)
// on locally available indices: e.g. h==8 -> 8 : 4 : 2.
// One disperse operation for every time we can divide h by 2.
void localDisperse ( in uint h )
{
	uint	t = gl_LocalInvocationID.x;

	for ( ; h > 1 ; h /= 2 ) 
	{
		barrier ();

		uint	halfH  = h >> 1;
		uint	t2     = t % halfH;
		uint	q      = h * ( ( 2 * t ) / h );

		localCompareAndSwap ( ivec2 ( q + t2, q + halfH + t2 ) );
	}
}

	// binary merge sort for local elements
void localSort ( in uint h )
{
	for ( uint hh = 2; hh <= h; hh <<= 1 ) 
	{
		localFlip     ( hh     );
		localDisperse ( hh / 2 );
	}
}

void	main ()
{
	uint t = gl_LocalInvocationID.x;
 
	 	// Calculate global offset for local workgroup
	uint offset = gl_WorkGroupSize.x * 2 * gl_WorkGroupID.x;

		// This shader can be called in four different modes:
		// 
		//  1. local flip+disperse (up to n == local_size_x * 2) 
		//  2. big flip
		//  3. big disperse
		//  4. local disperse 

	if ( algorithm <= LocalDisperse )
	{
			// for local store data in shared memory
		local_value [t*2    ] = value [offset + t*2    ];
		local_value [t*2 + 1] = value [offset + t*2 + 1];
	}

	switch ( algorithm )
	{
		case LocalBitonicMergeSort:
			localSort ( step );
			break;

		case LocalDisperse:
			localDisperse ( step );
			break;

		case BigFlip:
			bigFlip ( step );
			break;

		case BigDisperse:
			bigDisperse ( step );
			break;
	}

		// write shared data back to buffer
	if ( algorithm <= LocalDisperse )
	{
		barrier ();

			// push to global memory
		value [offset + t*2    ] = local_value [t*2    ];
		value [offset + t*2 + 1] = local_value [t*2 + 1];
	}
}